/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#include <Carbon/Carbon.h>

#ifndef ArmeArchivable_h
#include "ArmeArchivable.h"
#endif //ArmeArchivable_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

enum cSplineInterpolationType
{
	cSplineNoInterpolation = 0,
	cSplineLinearInterpolation,
	cSplineCubicInterpolation
};

class cSplineKnot
{
	ArmeTime					fTime;
	float						fValue;
	cSplineKnot*				fNext;
	cSplineKnot*				fPrevious;
	cSplineInterpolationType	fInterpolationType;
	
	float						fSplineCoeff[4];

public:
	cSplineKnot( ArmeTime t, float value);
	cSplineKnot();
	
	cSplineKnot*	nextKnot();
	cSplineKnot*	previousKnot();
	void			setNextKnot( cSplineKnot* next );
	void			setPreviousKnot( cSplineKnot* prev );

	cSplineInterpolationType interpolationType();
	void			setInterpolationType( cSplineInterpolationType type );

	float			value();
	ArmeTime		time();
	void			setValue( float value );
	void			setTime( ArmeTime time );
	
	void			setCoefficients( float a, float b, float c, float d );
	float*			getCoefficients();
};

enum cSplineEndCondition
{
	cSplineNaturalEnd = 0,
	cSplineLinearEnd,
	cSplineCubicEnd,
	cSplineDerivativeEnd
};

class cSpline
{
	bool					fCoeffsCalculated;
	cSplineKnot*			fFirstKnot;
	cSplineEndCondition		fEndCondition;
	int						fCount;

public:
	cSpline();
	~cSpline();
	
	cSplineEndCondition		endContidition();
	void					setEndCondition( cSplineEndCondition cond );
	cSplineKnot				*firstKnot();
	
	void					addKnot( cSplineKnot* newKnot, cSplineKnot *prevKnot );
	void					addKnot( cSplineKnot* newKnot);
	void					removeKnot( cSplineKnot* newKnot );
	void					moveKnot( cSplineKnot* movedKnot, cSplineKnot* newPreviousKnot);
	void					changeKnotValues( cSplineKnot *knot, ArmeTime newTime, float newValue );

	// returns the knot that with t-hTresh <= knot.time <= t+hTresh and a similary test for vertical
	cSplineKnot*			knotAround( ArmeTime t, float val, ArmeTime hTresh, float vTresh );

	void					computeCoeffs();
	float					splineValue( const ArmeTime t, bool &isRecorded );
};


/*
	cubic splines:
	
	piecewise cubic spline:
		for x[i-1] <= x <= x[i]
			Si(x)	= a[i] + b[i] * (x - x[i]) + c[i] * (x - x[i])^2 + d[i] * (x - x[i])^3
			S'i(x)	= b[i] - 2 * c[i] * (x - x[i]) + 3 * d[i] *(x - x[i])^2
			S"i(x)	= 2 * c[i] - 6 * d[i] * (x - x[i])
			
		conditions for a cubic spline:
		
		A: S(x[i]) = y[i]				// the cubic splines goes through the given points
		B: S[i](x[i]) = S[i+1](x[i])	// the spline is continuous (there are no "hops" at knots)
		C: S'[i](x[i]) = S'[i+1](x[i])	// the derivative at point x[i] is identical for the two pieces x[i] is a member of
		D: S"[i](x[i]) = S"[i+1](x[i])	// the second derivative at point x[i] is identical for the two pieces x[i] is a member of
		
			4n coefficents:
			A gives n+1 equations, B, C and D each give (n-1) equations
			extra condtions:
		E:		S"(x[0]) = S"(x[n]) = 0			(for natural cubic spline)
				
				
		let h[i] = ( x[i] - x[i-1] )
		
		Condition A gives:								S(x[i]) = y[i]
		
				S(x[i]) = a[i] + b[i] * 0 + c[i] * 0^2 + d[i] * 0^3 = y[i]
				
	(1)	-->		a[i] = y[i]
		
		
		Condition B gives:								S[i](x[i]) = S[i+1](x[i])
		
				S[i](x[i]) = a[i]
				S[i+1](x[i]) = a[i+1] + b[i+1] * (x[i] - x[i+1]) + c[i+1] * (x[i] - x[i+1])^2 + d[i+1] * (x[i] - x[i+1])^3
				
	(2)	-->		a[i] = a[i+1] - b[i+1] * h[i+1] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3
		
		
		Condition C gives:								S'[i](x[i]) = S'[i+1](x[i])
			
				S'[i](x[i])	= b[i] - 2 * c[i] * 0 + 3 * d[i] *0 = b[i]
				S'[i+1](x[i]) = b[i+1] - 2 * c[i+1] * (x[i] - x[i+1]) + 3 * d[i+1] *(x[i] - x[i+1])^2
				
	(3) -->		b[i] = b[i+1] - 2 * c[i+1] * h[i+1] + 3 * d[i+1] * h[i+1]^2

				
		Condition D gives:								S"[i](x[i]) = S"[i+1](x[i])
		
				S"[i](x[i]) 	= 2 * c[i] -6 * d[i] * 0
								= 2 * c[i]
				S"[i+1](x[i]) 	= 2 * c[i+1] - 6 * d[i+1] * (x[i] - x[i+1])
				
	(4) -->		2 * c[i] = 2 * c[i+1] - 6 * d[i+1] * h[i+1]
			

		From (4):
				
						2 * c[i-1] = 2 * c[i] - 6 * d[i] * h[i]
						d[i] = (2 * c[i-1] - 2 * c[i]) / ( -6 * h[i])

	(5) -->		d[i] = (c[i] - c[i-1])/ (3*h[i])


		From (E):										S"(x[0]) = S"(x[n]) = 0
		
	(6)	-->		d[1] = c[1]/3*h[1]
				c[n] = 0
						
		From (2):										a[i] = a[i+1] - b[i+1] * h[i+1] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3
	
				a[i]				= a[i+1] - b[i+1] * h[i+1] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3
				0 					= a[i+1] - a[i] - b[i+1] * h[i+1] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3
				b[i+1] * h[i+1] 	= a[i+1] - a[i] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3
						b[i+1]		= (a[i+1] - a[i] + c[i+1] * h[i+1]^2 - d[i+1] * h[i+1]^3) / h[i+1]

	(7) -->		b[i]		= (a[i] - a[i-1] + c[i] * h[i]^2 - d[i] * h[i]^3) / h[i]
		
					b[i+1] - b[i] = (a[i+1]-a[i]) / h[i+1] - (a[i]-a[i-1]) / h[i] + c[i+1]*h[i+1] - c[i]*h[i] + d[i+1] * h[i+1]^2 - d[i] * h[i]^2
					
		Modifying (3):
		
					b[i] = b[i+1] - 2 * c[i+1] * h[i+1] + 3 * d[i+1] * h[i+1]^2
					b[i] - b[i+1]  = - 2 * c[i+1] * h[i+1] + 3 * d[i+1] * h[i+1]^2

					b[i+1] - b[i]  = 2 * c[i+1] * h[i+1] - 3 * d[i+1] * h[i+1]^2
		Using (7)

				2 * c[i+1] * h[i+1] - 3 * d[i+1] * h[i+1]^2 = (a[i+1]-a[i]) / h[i+1] - (a[i]-a[i-1]) / h[i] + c[i+1]*h[i+1] - c[i]*h[i] + d[i+1] * h[i+1]^2 - d[i] * h[i]^2
				
		Using (1)
		
				2 * c[i+1] * h[i+1] - 3 * d[i+1] * h[i+1]^2 = (y[i+1]-y[i]) / h[i+1] - (y[i]-y[i-1]) / h[i] + c[i+1]*h[i+1] - c[i]*h[i] + d[i+1] * h[i+1]^2 - d[i] * h[i]^2

		Using (5)	d[i] = (c[i] - c[i-1])/ (3*h[i])
		
				2 * c[i+1] * h[i+1] - 3 * (c[i+1] - c[i])/ (3*h[i+1]) * h[i+1]^2 = (y[i+1]-y[i]) / h[i+1] - (y[i]-y[i-1]) / h[i] + c[i+1]*h[i+1] - c[i]*h[i] + d[i+1] * h[i+1]^2 - (c[i] - c[i-1])/ (3*h[i]) * h[i]^2
				
				2*c[i+1]*h[i+1] - 3*(c[i+1] - c[i])/(3*h[i+1])*h[i+1]^2 = (y[i+1]-y[i])/h[i+1] - (y[i]-y[i-1])/h[i] + c[i+1]*h[i+1] - c[i]*h[i] + ((c[i+1] - c[i])/(3*h[i+1]))*h[i+1]^2 - ((c[i] - c[i-1])/(3*h[i]))*h[i]^2
			
				2*c[i+1]*h[i+1] - (c[i+1] - c[i])*h[i+1] = (y[i+1]-y[i])/h[i+1] - (y[i]-y[i-1])/h[i] + c[i+1]*h[i+1] - c[i]*h[i] + ((c[i+1] - c[i])*h[i+1]/3) - ((c[i] - c[i-1])*h[i]/(3))
				
				+ 2*c[i]*h[i+1]
				+ 2*c[i]*h[i]
				+ c[i+1]*h[i+1]
				+ c[i-1]*h[i]
				= 
				3 * ( (y[i+1]-y[i])/h[i+1] - (y[i]-y[i-1])/h[i] )

	(8) -->		h[i] * c[i-1] + 2*c[i]*(h[i]+h[i+1]) + h[i+1]*c[i+1] = 3 * ( (y[i+1]-y[i])/h[i+1] - (y[i]-y[i-1])/h[i] )


		(E) and (6) give:
				
	(9) -->		b[1] = (y[1] - y[0])/ h[1] + 2*h[1]*c[1]/3
	
		(9), (3) and (5) give:
				
	(10) -->	2 *( h[1] + h[2])* c[1] + h[2]* c[2] = 3 ( (y[2] - y[1])/h[2] - (y[1] - y[0])/h[1] )

	From (8) and (10):
	
		if 	k = 2 * ( h[i] + h[i+1])		
		and g = 3 * ( (y[i+1] - y[i])/h[i+1] - (y[i] - y[i-1])/h[i])
	
	+------------------------------------------------+		+------+		+------+
	|  k1     h2      0     ...     0      0     0   |		|  c1  |		|  g1  |
	|												 |		|	   |		|	   |
	|  h2     k2     h3     ...     0      0     0   |		|  c2  |		|  g2  |
	|												 |		|	   |		|	   |
	|  0      h3     k3     ...     0      0     0   |		|  c3  |		|  g3  |
	|												 |		|	   |		|	   |
	|  :       :      :     :.      :      :     :   |	*	|  :   |   =	|   :  |
	|												 |		|	   |		|	   |
	|  0       0      0     ...  k[n-3] h[n-2]   0   |		|c[n-3]|		|g[n-3]|
	|												 |		|	   |		|	   |
	|  0       0      0     ...  h[n-2] k[n-2] h[n-1]| 		|c[n-2]|		|g[n-2]|
	|												 |		|	   |		|	   |
	|  0       0      0     ...     0   h[n-1] k[n-1]|		|c[n-1]|		|g[n-1]|
	+------------------------------------------------+		+------+		+------+
			
*/
